from .errors.parsing import Parsing
from .constants import (
    EMPTY_ELEMENT,
    FIELD,
    FIELDSET,
    LIST,
    MULTILINE_FIELD_BEGIN,
    SECTION
)

class Resolver:
    def __init__(self, context):
        self._context = context

        self._unresolved_sections = hasattr(self._context, '_copy_sections')
        self._unresolved_non_section_elements = hasattr(self._context, '_copy_non_section_elements')

    def consolidate_non_section_elements(self, element, template):
        # TODO: Do this lazily as todo'd elsewhere
        if 'comments' in template and not 'comments' in element:
            element['comments'] = template['comments']

        if element['type'] == EMPTY_ELEMENT:
            if template['type'] == MULTILINE_FIELD_BEGIN:
                element['type'] = FIELD  # TODO: Revisit this - maybe should be MULTILINE_FIELD_COPY or something else - consider implications all around.
                self.mirror(element, template)
            elif template['type'] == FIELD:
                element['type'] = FIELD
                self.mirror(element, template)
            elif template['type'] == FIELDSET:
                element['type'] = FIELDSET
                self.mirror(element, template)
            elif template['type'] == LIST:
                element['type'] = LIST
                self.mirror(element, template)
        elif element['type'] == FIELDSET:
            if template['type'] == FIELDSET:
                element['extend'] = template
            elif (template['type'] == FIELD or
                  template['type'] == LIST or
                  template['type'] == MULTILINE_FIELD_BEGIN):
                raise Parsing.missing_fieldset_for_fieldset_entry(self._context, element['entries'][0])
        elif element['type'] == LIST:
            if template['type'] == LIST:
                element['extend'] = template
            elif (template['type'] == FIELD or
                  template['type'] == FIELDSET or
                  template['type'] == MULTILINE_FIELD_BEGIN):
                raise Parsing.missing_list_for_list_item(self._context, element['items'][0])

    def consolidate_sections(self, section, template, deep_merge):
        if 'comments' in template and not 'comments' in section:
            section['comments'] = template['comments']

        if len(section['elements']) == 0:
            self.mirror(section, template)
        else:
            # TODO: Handle possibility in two templates (one hardcoded in the document, one implicitly derived through deep merging)
            #       Possibly also elswhere (e.g. up there in the mirror branch?)
            section['extend'] = template

            if not deep_merge:
                return

            merge_map = {}

            for section_element in section['elements']:
                if section_element['type'] != SECTION or section_element['key'] in merge_map:
                    merge_map[section_element['key']] = False # non-mergable (no section or multiple sections with same key)
                else:
                    merge_map[section_element['key']] = { 'section': section_element }

            for section_element in template['elements']:
                if section_element['key'] in merge_map:
                    merger = merge_map[section_element['key']]

                    if merger is False:
                        continue

                    if section_element['type'] != SECTION or 'template' in merger:
                        merge_map[section_element['key']] = False # non-mergable (no section or multiple template sections with same key)
                    else:
                        merger['template'] = section_element

            for merger in merge_map.values():
                if merger and 'template' in merger:
                    self.consolidate_sections(merger['section'], merger['template'], True)

    def mirror(self, element, template):
        if 'mirror' in template:
            element['mirror'] = template['mirror']
        else:
            element['mirror'] = template

    def resolve_non_section_element(self, element, previous_elements):
        if element in previous_elements:
            raise Parsing.cyclic_dependency(self._context, element, previous_elements)

        template = element['copy']['template']

        if 'copy' in template: # TODO: Maybe we change that to .unresolved everywhere ?
            self.resolve_non_section_element(template, [*previous_elements, element])

        self.consolidate_non_section_elements(element, template)

        del element['copy']

    def resolve_section(self, section, previous_sections):
        if section in previous_sections:
            raise Parsing.cyclic_dependency(self._context, section, previous_sections)

        if 'deep_resolve' in section:
            for section_element in section['elements']:
                if section_element['type'] == SECTION and ('copy' in section_element or 'deep_resolve' in section_element):
                    self.resolve_section(section_element, [*previous_sections, section])

            del section['deep_resolve']

        if 'copy' in section:
            template = section['copy']['template']

            if 'copy'in template or 'deep_resolve' in template:
                self.resolve_section(template, [*previous_sections, section])

            self.consolidate_sections(section, template, 'deep_copy' in section)

            del section['copy']

    def index(self, section):
        for element in section['elements']:
            if element['type'] == SECTION:
                self.index(element)

                if (self._unresolved_sections and
                    element['key'] in self._context._copy_sections and
                    ('template' not in element or element['key'] != element['template'])):
                    copy_data = self._context._copy_sections[element['key']]

                    if 'template' in copy_data:
                        raise Parsing.two_or_more_templates_found(self._context, copy_data['targets'][0], copy_data['template'], element)

                    copy_data['template'] = element
            elif (self._unresolved_non_section_elements and
                  element['key'] in self._context._copy_non_section_elements and
                  ('template' not in element or element['key'] != element['template'])):
                copy_data = self._context._copy_non_section_elements[element['key']]

                if 'template' in copy_data:
                    raise Parsing.two_or_more_templates_found(self._context, copy_data['targets'][0], copy_data['template'], element)

                copy_data['template'] = element

    def resolve(self):
        self.index(self._context.document)

        if self._unresolved_non_section_elements:
            for copy in self._context._copy_non_section_elements.values():
                if not 'template' in copy:
                    raise Parsing.non_section_element_not_found(self._context, copy['targets'][0])

                for target in copy['targets']:
                    if 'copy' in target:
                        self.resolve_non_section_element(target, [])

            del self._context._copy_non_section_elements

        if self._unresolved_sections:
            for copy in self._context._copy_sections.values():
                if not 'template' in copy:
                    raise Parsing.section_not_found(self._context, copy['targets'][0])

                for target in copy['targets']:
                    if 'copy' in target:
                        self.resolve_section(target, [])

            del self._context._copy_sections



#  GENERATED ON 2019-05-30T07:54:15 - DO NOT EDIT MANUALLY

content_header = 'Inhalt'
expected_document = 'Das Dokument wurde erwartet.'
expected_empty = 'Ein leeres Element wurde erwartet.'
expected_field = 'Ein Feld wurde erwartet.'
expected_fields = 'Nur Felder wurden erwartet.'
expected_fieldset = 'Ein Fieldset wurde erwartet.'
expected_fieldset_entry = 'Ein Fieldset Eintrag wurde erwartet.'
expected_fieldsets = 'Nur Fieldsets wurden erwartet.'
expected_list = 'Eine Liste wurde erwartet.'
expected_list_item = 'Ein Listen Eintrag wurde erwartet.'
expected_lists = 'Nur Listen wurden erwartet.'
expected_section = 'Eine Sektion wurde erwartet.'
expected_sections = 'Nur Sektionen wurden erwartet.'
expected_single_element = 'Nur ein einzelnes Element wurde erwartet.'
expected_single_empty = 'Nur ein einzelnes leeres Element wurde erwartet.'
expected_single_field = 'Nur ein einzelnes Feld wurde erwartet.'
expected_single_fieldset = 'Nur ein einzelnes Fieldset wurde erwartet.'
expected_single_fieldset_entry = 'Nur ein einzelner Fieldset Eintrag wurde erwartet.'
expected_single_list = 'Nur eine einzelne Liste wurde erwartet.'
expected_single_section = 'Nur eine einzelne Sektion wurde erwartet.'
gutter_header = 'Zeile'
missing_comment = 'Ein erforderlicher Kommentar zu diesem Feld fehlt.'
missing_element = 'Ein einzelnes Element ist erforderlich - es kann einen beliebigen Schlüssel haben.'
missing_empty = 'Ein einzelnes leeres Element ist erforderlich - es kann einen beliebigen Schlüssel haben.'
missing_field = 'Ein einzelnes Feld ist erforderlich - es kann einen beliebigen Schlüssel haben.'
missing_fieldset = 'Ein einzelnes Fieldset ist erforderlich - es kann einen beliebigen Schlüssel haben.'
missing_fieldset_entry = 'Ein einzelner Fieldset Eintrag ist erforderlich - er kann einen beliebigen Schlüssel haben.'
missing_list = 'Eine einzelne Liste ist erforderlich - sie kann einen beliebigen Schlüssel haben.'
missing_section = 'Eine einzelne Sektion ist erforderlich - sie kann einen beliebigen Schlüssel haben.'
unexpected_element = 'Dieses Element wurde nicht erwartet, prüfe ob es am richtigen Platz ist und dass der Schlüssel keine Tippfehler enthält.'
comment_error = lambda message: f"Es gibt ein Problem mit dem Kommentar dieses Elements: {message}"
cyclic_dependency = lambda line, key: f"In Zeile {line} wird '{key}' in sich selbst kopiert."
expected_empty_with_key = lambda key: f"Ein leeres Element mit dem Schlüssel '{key}' wurde erwartet."
expected_field_with_key = lambda key: f"Ein Feld mit dem Schlüssel '{key}' wurde erwartet."
expected_fields_with_key = lambda key: f"Nur Felder mit dem Schlüssel '{key}' wurden erwartet."
expected_fieldset_with_key = lambda key: f"Ein Fieldset mit dem Schlüssel '{key}' wurde erwartet."
expected_fieldsets_with_key = lambda key: f"Nur Fieldsets mit dem Schlüssel '{key}' wurden erwartet."
expected_list_with_key = lambda key: f"Eine Liste mit dem Schlüssel '{key}' wurde erwartet."
expected_lists_with_key = lambda key: f"Nur Listen mit dem Schlüssel '{key}' wurden erwartet."
expected_section_with_key = lambda key: f"Eine Sektion mit dem Schlüssel '{key}' wurde erwartet."
expected_sections_with_key = lambda key: f"Nur Sektionen mit dem Schlüssel '{key}' wurden erwartet."
expected_single_element_with_key = lambda key: f"Nur ein einzelnes Element mit dem Schlüssel '{key}' wurde erwartet."
expected_single_empty_with_key = lambda key: f"Nur ein einzelnes leeres Element mit dem Schlüssel '{key}' wurde erwartet."
expected_single_field_with_key = lambda key: f"Nur ein einzelnes Feld mit dem Schlüssel '{key}' wurde erwartet."
expected_single_fieldset_entry_with_key = lambda key: f"Nur ein einzelner Fieldset Eintrag mit dem Schlüssel '{key}' wurde erwartet."
expected_single_fieldset_with_key = lambda key: f"Nur ein einzelnes Fieldset mit dem Schlüssel '{key}' wurde erwartet."
expected_single_list_with_key = lambda key: f"Nur eine einzelne Liste mit dem Schlüssel '{key}' wurde erwartet."
expected_single_section_with_key = lambda key: f"Nur eine einzelne Sektion mit dem Schlüssel '{key}' wurde erwartet."
invalid_line = lambda line: f"Zeile {line} folgt keinem spezifierten Muster."
key_error = lambda message: f"Es gibt ein Problem mit dem Schlüssel dieses Elements: {message}"
missing_element_for_continuation = lambda line: f"Zeile {line} enthält eine Zeilenfortsetzung ohne dass davor ein fortsetzbares Element begonnen wurde."
missing_element_with_key = lambda key: f"Das Element '{key}' fehlt - falls es angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
missing_empty_with_key = lambda key: f"Das leere Element '{key}' fehlt - falls es angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
missing_field_value = lambda key: f"Das Feld '{key}' muss einen Wert enthalten."
missing_field_with_key = lambda key: f"Das Feld '{key}' fehlt - falls es angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
missing_fieldset_entry_value = lambda key: f"Der Fieldset Eintrag '{key}' muss einen Wert enthalten."
missing_fieldset_entry_with_key = lambda key: f"Der Fieldset Eintrag '{key}' fehlt - falls er angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
missing_fieldset_for_fieldset_entry = lambda line: f"Zeile {line} enthält einen Fieldset Eintrag ohne dass davor ein Fieldset begonnen wurde."
missing_fieldset_with_key = lambda key: f"Das Fieldset '{key}' fehlt - falls es angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
missing_list_for_list_item = lambda line: f"Zeile {line} enthält einen Listeneintrag ohne dass davor ein eine Liste begonnen wurde."
missing_list_item_value = lambda key: f"Die Liste '{key}' darf keine leeren Einträge enthalten."
missing_list_with_key = lambda key: f"Die Liste '{key}' fehlt - falls sie angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
missing_section_with_key = lambda key: f"Die Sektion '{key}' fehlt - falls sie angegeben wurde eventuell nach Tippfehlern Ausschau halten und auch die Gross/Kleinschreibung beachten."
non_section_element_not_found = lambda line, key: f"In Zeile {line} soll das Nicht-Sektions Element '{key}' kopiert werden, es wurde aber nicht gefunden."
section_hierarchy_layer_skip = lambda line: f"Zeile {line} beginnt eine Sektion die mehr als eine Ebene tiefer liegt als die aktuelle."
section_not_found = lambda line, key: f"In Zeile {line} soll die Sektion '{key}' kopiert werden, sie wurde aber nicht gefunden."
two_or_more_templates_found = lambda key: f"Es gibt mindestens zwei Elemente mit dem Schlüssel '{key}' die hier zum kopieren in Frage kommen, es ist nicht klar welches kopiert werden soll."
unterminated_escaped_key = lambda line: f"In Zeile {line} wird der Schlüssel eines Elements escaped, jedoch wird diese Escape Sequenz bis zum Ende der Zeile nicht mehr beendet."
unterminated_multiline_field = lambda key, line: f"Das Mehrzeilenfeld '{key}' dass in Zeile {line} beginnt wird bis zum Ende des Dokuments nicht mehr beendet."
value_error = lambda message: f"Es gibt ein Problem mit dem Wert dieses Elements: {message}"


#  GENERATED ON 2019-05-30T07:54:15 - DO NOT EDIT MANUALLY

content_header = 'Contenido'
expected_document = 'Se esperaba el documento.'
expected_empty = 'Se esperaba un elemento vacío.'
expected_field = 'Se esperaba una casilla.'
expected_fields = 'Solo se esperaban casillas.'
expected_fieldset = 'Se esperaba una collecíon de casillas.'
expected_fieldset_entry = 'Se esperaba una casilla de collecíon.'
expected_fieldsets = 'Solo se esperaban colleciones de casillas.'
expected_list = 'Se esperaba una lista.'
expected_list_item = 'Se esperaba una entrada de lista.'
expected_lists = 'Solo se esperaban listas.'
expected_section = 'Se esperaba una sección.'
expected_sections = 'Solo se esperaban secciones.'
expected_single_element = 'Solo se esperaba un único elemento.'
expected_single_empty = 'Solo se esperaba un único elemento vacío.'
expected_single_field = 'Solo se esperaba una sola casilla.'
expected_single_fieldset = 'Solo se esperaba una sola collecíon de casillas.'
expected_single_fieldset_entry = 'Solo se esperaba una sola casilla de collecíon.'
expected_single_list = 'Solo se esperaba una sola lista.'
expected_single_section = 'Solo se esperaba una sola sección.'
gutter_header = 'Línea'
missing_comment = 'Falta un comentario necesario para este elemento.'
missing_element = 'Se requiere un único elemento - puede tener cualquier clave.'
missing_empty = 'Se requiere un único elemento vacío - puede tener cualquier clave.'
missing_field = 'Se requiere una sola casilla - puede tener cualquier clave.'
missing_fieldset = 'Se requiere una sola collecíon de casillas - puede tener cualquier clave.'
missing_fieldset_entry = 'Se requiere una sola casilla de collecíon - puede tener cualquier clave.'
missing_list = 'Se requiere una sola lista - puede tener cualquier clave.'
missing_section = 'Se requiere una sola sección - puede tener cualquier clave.'
unexpected_element = 'Este elemento no se esperaba, averigua si es en el sitio correcto y que no contiene un error tipográfico la clave.'
comment_error = lambda message: f"Hay un problema con el comentario de este elemento: {message}"
cyclic_dependency = lambda line, key: f"En la línea {line} '{key}' se copia en sí mismo."
expected_empty_with_key = lambda key: f"Se esperaba un elemento vacío con la clave '{key}'."
expected_field_with_key = lambda key: f"Se esperaba una casilla con la clave '{key}'."
expected_fields_with_key = lambda key: f"Solo se esperaban casillas con la clave '{key}'."
expected_fieldset_with_key = lambda key: f"Se esperaba una collecíon de casillas con la clave '{key}'."
expected_fieldsets_with_key = lambda key: f"Solo se esperaban colleciones de casillas con la clave '{key}'."
expected_list_with_key = lambda key: f"Se esperaba una lista con la clave '{key}'."
expected_lists_with_key = lambda key: f"Solo se esperaban listas con la clave '{key}'."
expected_section_with_key = lambda key: f"Se esperaba una sección con la clave '{key}'."
expected_sections_with_key = lambda key: f"Solo se esperaban secciones con la clave '{key}'."
expected_single_element_with_key = lambda key: f"Solo se esperaba un único elemento con la clave '{key}'."
expected_single_empty_with_key = lambda key: f"Solo se esperaba un único elemento vacío con la clave '{key}'."
expected_single_field_with_key = lambda key: f"Solo se esperaba una sola casilla con la clave '{key}'."
expected_single_fieldset_entry_with_key = lambda key: f"Solo se esperaba una sola casilla de collecíon con la clave '{key}'."
expected_single_fieldset_with_key = lambda key: f"Solo se esperaba una sola collecíon de casillas con la clave '{key}'."
expected_single_list_with_key = lambda key: f"Solo se esperaba una sola lista con la clave '{key}'."
expected_single_section_with_key = lambda key: f"Solo se esperaba una sola sección con la clave '{key}'."
invalid_line = lambda line: f"Línea {line} no sigue un patrón especificado."
key_error = lambda message: f"Hay un problema con la clave de este elemento: {message}"
missing_element_for_continuation = lambda line: f"Línea {line} contiene una continuacíon de línea sin un elemento que se puede continuar empezado antes."
missing_element_with_key = lambda key: f"Falta el elemento '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
missing_empty_with_key = lambda key: f"Falta el elemento vacío '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
missing_field_value = lambda key: f"La casilla '{key}' debe contener un valor."
missing_field_with_key = lambda key: f"Falta la casilla '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
missing_fieldset_entry_value = lambda key: f"La casilla de collecíon '{key}' debe contener un valor."
missing_fieldset_entry_with_key = lambda key: f"Falta la casilla de collecíon '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
missing_fieldset_for_fieldset_entry = lambda line: f"Línea {line} contiene una casilla de collecíon sin una collecíon de casillas empezada antes."
missing_fieldset_with_key = lambda key: f"Falta la collecíon de casillas '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
missing_list_for_list_item = lambda line: f"Línea {line} contiene una entrada de lista sin una lista empezada antes."
missing_list_item_value = lambda key: f"La lista '{key}' no debe contener entradas vacías."
missing_list_with_key = lambda key: f"Falta la lista '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
missing_section_with_key = lambda key: f"Falta la sección '{key}' - si se proporcionó, mira por errores ortográficos y también distingue entre mayúsculas y minúsculas."
non_section_element_not_found = lambda line, key: f"En la línea {line} debe ser copiado el elemento no sección '{key}', pero no se encontró."
section_hierarchy_layer_skip = lambda line: f"Línea {line} inicia una sección que es más de un nivel más bajo el actual."
section_not_found = lambda line, key: f"En la línea {line} debe ser copiado la sección '{key}', pero no se encontró."
two_or_more_templates_found = lambda key: f"Hay como mínimo dos elementos con la clave '{key}' que se clasifiquen para estar copiado, no está claro cual debe ser copiado."
unterminated_escaped_key = lambda line: f"En la línea {line}, la clave de un elemento se escapa, pero esta secuencia de escape no termina hasta el final de la línea."
unterminated_multiline_field = lambda key, line: f"La casilla de múltiples líneas '{key}' que comienza en la línea {line} no termina hasta el final del documento."
value_error = lambda message: f"Hay un problema con el valor de este elemento: {message}"


from . import list_item
from . import section
from .element_base import ElementBase


class List(ElementBase):
    def __repr__(self):
        return f"<class List key={self._instruction['key']} items={len(self._items())}>"

    def _instantiate_items(self, list):
        if 'mirror' in list:
            return self._instantiate_items(list['mirror'])
        elif 'extend' in list:
            return self._instantiate_items(list['extend']) + [list_item.ListItem(self._context, item, self) for item in list['items']]
        elif 'items' in list:
            return [list_item.ListItem(self._context, item, self) for item in list['items']]
        else:
            return []

    def _items(self):
        if not hasattr(self, '_instantiated_items'):
            self._instantiated_items = self._instantiate_items(self._instruction)

        return self._instantiated_items

    def _untouched(self):
        if not hasattr(self, '_touched'):
            return self._instruction

        untouched_item = next((item for item in self._items() if not hasattr(item, '_touched')), None)

        return untouched_item._instruction if untouched_item else False

    def items(self):
        self._touched = True

        return self._items()

    def length(self):
        self._touched = True

        return len(self._items())

    def optional_string_values(self):
        self._touched = True

        return [item.optional_string_value() for item in self._items()]

    def optional_values(self, loader):
        self._touched = True

        return [item.optional_value(loader) for item in self._items()]

    def parent(self):
        return self._parent or section.Section(self._context, self._instruction['parent'])

    def required_string_values(self):
        self._touched = True

        return [item.required_string_value() for item in self._items()]

    def required_values(self, loader):
        self._touched = True

        return [item.required_value(loader) for item in self._items()]

    def touch(self):
        self._touched = True

        for item in self._items():
            item._touched = True



from .missing_value_element_base import MissingValueElementBase

class MissingFieldsetEntry(MissingValueElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingFieldsetEntry key={self._key}>"

        return '<class MissingFieldsetEntry>'



from ..errors.validation import Validation
from ..constants import DOCUMENT, LIST_ITEM

class ElementBase:
    def __init__(self, context, instruction, parent=None):
        self._context = context
        self._instruction = instruction
        self._parent = parent  # TODO: Could be set optionally as well (needs hasattr checks everywhere then)

    def _comment(self, loader=None, *, required):
        self._touched = True

        comment = self._context.comment(self._instruction)

        if not comment:
            if required:
                raise Validation.missing_comment(self._context, self._instruction)
            else:
                return None

        if not loader:
            return comment

        try:
            return loader(comment)
        except ValueError as message:
            raise Validation.comment_error(self._context, message, self._instruction)

    def _key(self):
        if self._instruction['type'] is DOCUMENT:
            return None

        if self._instruction['type'] is LIST_ITEM:
            return self._instruction['parent']['key']

        return self._instruction['key']

    def comment_error(self, message):
        if callable(message):
            message = message(self._context.comment(self._instruction))

        return Validation.comment_error(self._context, message, self._instruction)

    def error(self, message):
        if callable(message):
            message = message(self)  # Revisit self in this context - problematic

        return Validation.element_error(self._context, message, self._instruction)

    def key(self, loader):
        self._touched = True

        try:
            return loader(self._key())
        except ValueError as message:
            raise Validation.key_error(self._context, message, self._instruction)

    def key_error(self, message):
        if callable(message):
            message = message(self._key())

        return Validation.key_error(self._context, message, self._instruction)

    def optional_comment(self, loader):
        return self._comment(loader, required=False)

    def optional_string_comment(self):
        return self._comment(required=False)

    def raw(self):
        return self._context.raw(self._instruction)

    def required_comment(self, loader):
        return self._comment(loader, required=True)

    def required_string_comment(self):
        return self._comment(required=True)

    def string_key(self):
        self._touched = True

        return self._key()

    def touch(self):
        self._touched = True



class MissingElementBase:
    def __init__(self, key, parent):
        self._key = key
        self._parent = parent

    def _missing_error(self, _element):
        self._parent._missing_error(self)

    def key(self, _loader):
        self._parent._missing_error(self)

    def optional_comment(self, _loader):
        return None

    def optional_string_comment(self):
        return None

    def raw(self):
        return None

    def required_comment(self, _loader):
        self._parent._missing_error(self)

    def required_string_comment(self):
        self._parent._missing_error(self)

    def string_key(self):
        self._parent._missing_error(self)



from .missing_value_element_base import MissingValueElementBase

class MissingField(MissingValueElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingField key={self._key}>"

        return '<class MissingField>'



import re
from .errors.parsing import Parsing
from .grammar_regex import Grammar
from .constants import (
    BEGIN,
    COMMENT,
    CONTINUATION,
    DOCUMENT,
    EMPTY_ELEMENT,
    END,
    FIELD,
    FIELDSET,
    FIELDSET_ENTRY,
    LIST,
    LIST_ITEM,
    MULTILINE_FIELD_BEGIN,
    MULTILINE_FIELD_END,
    MULTILINE_FIELD_VALUE,
    SECTION,
    UNPARSED
)

class Analyzer:
    def __init__(self, context):
        self._context = context
        self._depth = 0 # TODO: Make depth only a local temporary variable in analysis in js implementation too
        self._index = 0
        self._line = 0

    def _tokenize_error_context(self):
        first_instruction = None

        while True:
            end_of_line_index = self._context.input.find('\n', self._index)

            if end_of_line_index == -1:
                instruction = {
                    'ranges': { 'line': (self._index, len(self._context.input)) },
                    'line': self._line
                }

                if first_instruction is not None:
                    instruction['type'] = UNPARSED

                self._context.line_count = self._line + 1
                self._context.meta.append(instruction)

                return first_instruction or instruction
            else:
                instruction = {
                    'ranges': { 'line': (self._index, end_of_line_index) },
                    'line': self._line
                }

                self._context.meta.append(instruction)

                if first_instruction is None:
                    first_instruction = instruction
                else:
                    instruction['type'] = UNPARSED

                self._index = end_of_line_index + 1
                self._line += 1

    def analyze(self):
        if len(self._context.input) == 0:
            self._context.line_count = 1
            return

        comments = None
        last_continuable_element = None
        last_non_section_element = None
        last_section = self._context.document

        while self._index < len(self._context.input):
            match = Grammar.REGEX.search(self._context.input, self._index)

            if not match or match.start() != self._index:
                instruction = self._tokenize_error_context()
                raise Parsing.invalid_line(self._context, instruction)

            instruction = {
                'line': self._line,
                'ranges': {
                    'line': (self._index, match.end())
                }
            }

            multiline_field = False

            if match.group(Grammar.EMPTY_LINE_INDEX) is not None:

                if comments is not None:
                    self._context.meta.extend(comments)
                    comments = None

            elif match.group(Grammar.ELEMENT_OPERATOR_INDEX) is not None:

                if comments is not None:
                    instruction['comments'] = comments
                    comments = None

                instruction['key'] = match.group(Grammar.KEY_UNESCAPED_INDEX)

                if instruction['key'] is None:
                    instruction['key'] = match.group(Grammar.KEY_ESCAPED_INDEX)
                    instruction['ranges']['element_operator'] = match.span(Grammar.ELEMENT_OPERATOR_INDEX)
                    instruction['ranges']['escape_begin_operator'] = match.span(Grammar.KEY_ESCAPE_BEGIN_OPERATOR_INDEX)
                    instruction['ranges']['escape_end_operator'] = match.span(Grammar.KEY_ESCAPE_END_OPERATOR_INDEX)
                    instruction['ranges']['key'] = match.span(Grammar.KEY_ESCAPED_INDEX)
                else:
                    instruction['ranges']['element_operator'] = match.span(Grammar.ELEMENT_OPERATOR_INDEX)
                    instruction['ranges']['key'] = match.span(Grammar.KEY_UNESCAPED_INDEX)

                value = match.group(Grammar.FIELD_VALUE_INDEX)

                if value is None:
                    instruction['type'] = EMPTY_ELEMENT
                else:
                    instruction['type'] = FIELD
                    instruction['value'] = value
                    instruction['ranges']['value'] = match.span(Grammar.FIELD_VALUE_INDEX)

                instruction['parent'] = last_section
                last_section['elements'].append(instruction)
                last_continuable_element = instruction
                last_non_section_element = instruction

            elif match.group(Grammar.LIST_ITEM_OPERATOR_INDEX) is not None:

                if comments is not None:
                    instruction['comments'] = comments
                    comments = None

                instruction['ranges']['item_operator'] = match.span(Grammar.LIST_ITEM_OPERATOR_INDEX)
                instruction['type'] = LIST_ITEM

                value = match.group(Grammar.LIST_ITEM_VALUE_INDEX)

                if value is not None:
                    instruction['ranges']['value'] = match.span(Grammar.LIST_ITEM_VALUE_INDEX)
                    instruction['value'] = value

                if last_non_section_element is None:
                    instruction = self._tokenize_error_context()
                    raise Parsing.missing_list_for_list_item(self._context, instruction)
                elif last_non_section_element['type'] == LIST:
                    last_non_section_element['items'].append(instruction)
                elif last_non_section_element['type'] == EMPTY_ELEMENT:
                    last_non_section_element['items'] = [instruction]
                    last_non_section_element['type'] = LIST
                else:
                    instruction = self._tokenize_error_context()
                    raise Parsing.missing_list_for_list_item(self._context, instruction)

                instruction['parent'] = last_non_section_element
                last_continuable_element = instruction

            elif match.group(Grammar.FIELDSET_ENTRY_OPERATOR_INDEX):

                if comments is not None:
                    instruction['comments'] = comments
                    comments = None

                instruction['type'] = FIELDSET_ENTRY

                unescaped_key = match.group(Grammar.KEY_UNESCAPED_INDEX)

                if unescaped_key is None:
                    instruction['key'] = match.group(Grammar.KEY_ESCAPED_INDEX)
                    instruction['ranges']['escape_begin_operator'] = match.span(Grammar.KEY_ESCAPE_BEGIN_OPERATOR_INDEX)
                    instruction['ranges']['escape_end_operator'] = match.span(Grammar.KEY_ESCAPE_END_OPERATOR_INDEX)
                    instruction['ranges']['entry_operator'] = match.span(Grammar.FIELDSET_ENTRY_OPERATOR_INDEX)
                    instruction['ranges']['key'] = match.span(Grammar.KEY_ESCAPED_INDEX)
                else:
                    instruction['key'] = unescaped_key
                    instruction['ranges']['entry_operator'] = match.span(Grammar.FIELDSET_ENTRY_OPERATOR_INDEX)
                    instruction['ranges']['key'] = match.span(Grammar.KEY_UNESCAPED_INDEX)

                value = match.group(Grammar.FIELDSET_ENTRY_VALUE_INDEX)

                if value is not None:
                    instruction['ranges']['value'] = match.span(Grammar.FIELDSET_ENTRY_VALUE_INDEX)
                    instruction['value'] = value

                if last_non_section_element is None:
                    instruction = self._tokenize_error_context()
                    raise Parsing.missing_fieldset_for_fieldset_entry(self._context, instruction)
                elif last_non_section_element['type'] == FIELDSET:
                    last_non_section_element['entries'].append(instruction)
                elif last_non_section_element['type'] == EMPTY_ELEMENT:
                    last_non_section_element['entries'] = [instruction]
                    last_non_section_element['type'] = FIELDSET
                else:
                    instruction = self._tokenize_error_context()
                    raise Parsing.missing_fieldset_for_fieldset_entry(self._context, instruction)

                instruction['parent'] = last_non_section_element
                last_continuable_element = instruction

            elif match.group(Grammar.SPACED_LINE_CONTINUATION_OPERATOR_INDEX) is not None:

                if last_continuable_element is None:
                    instruction = self._tokenize_error_context()
                    raise Parsing.missing_element_for_continuation(self._context, instruction)

                instruction['spaced'] = True
                instruction['ranges']['spaced_line_continuation_operator'] = match.span(Grammar.SPACED_LINE_CONTINUATION_OPERATOR_INDEX)
                instruction['type'] = CONTINUATION

                value = match.group(Grammar.SPACED_LINE_CONTINUATION_VALUE_INDEX)

                if value is not None:
                    instruction['ranges']['value'] = match.span(Grammar.SPACED_LINE_CONTINUATION_VALUE_INDEX)
                    instruction['value'] = value

                if 'continuations' in last_continuable_element:
                    last_continuable_element['continuations'].append(instruction)
                else:
                    if last_continuable_element['type'] == EMPTY_ELEMENT:
                        last_continuable_element['type'] = FIELD

                    last_continuable_element['continuations'] = [instruction]

                if comments is not None:
                    self._context.meta.extend(comments)
                    comments = None

            elif match.group(Grammar.DIRECT_LINE_CONTINUATION_OPERATOR_INDEX) is not None:

                if last_continuable_element is None:
                    instruction = self._tokenize_error_context()
                    raise Parsing.missing_element_for_continuation(self._context, instruction)

                instruction['ranges']['direct_line_continuation_operator'] = match.span(Grammar.DIRECT_LINE_CONTINUATION_OPERATOR_INDEX)
                instruction['type'] = CONTINUATION

                value = match.group(Grammar.DIRECT_LINE_CONTINUATION_VALUE_INDEX)

                if value is not None:
                    instruction['ranges']['value'] = match.span(Grammar.DIRECT_LINE_CONTINUATION_VALUE_INDEX)
                    instruction['value'] = value

                if 'continuations' in last_continuable_element:
                    last_continuable_element['continuations'].append(instruction)
                else:
                    if last_continuable_element['type'] == EMPTY_ELEMENT:
                        last_continuable_element['type'] = FIELD

                    last_continuable_element['continuations'] = [instruction]

                if comments is not None:
                    self._context.meta.extend(comments)
                    comments = None

            elif match.group(Grammar.SECTION_OPERATOR_INDEX) is not None:

                if comments is not None:
                    instruction['comments'] = comments
                    comments = None

                instruction['elements'] = []
                instruction['ranges']['section_operator'] = match.span(Grammar.SECTION_OPERATOR_INDEX)
                instruction['type'] = SECTION

                new_depth = instruction['ranges']['section_operator'][END] - instruction['ranges']['section_operator'][BEGIN]

                if new_depth == self._depth + 1:
                    instruction['parent'] = last_section
                    self._depth = new_depth
                elif new_depth == self._depth:
                    instruction['parent'] = last_section['parent']
                elif new_depth < self._depth:
                    while new_depth < self._depth:
                        last_section = last_section['parent']
                        self._depth -= 1

                    instruction['parent'] = last_section['parent']
                else:
                    instruction = self._tokenize_error_context()
                    raise Parsing.section_hierarchy_layer_skip(self._context, instruction, last_section)

                instruction['parent']['elements'].append(instruction)
                last_section = instruction

                instruction['key'] = match.group(Grammar.SECTION_KEY_UNESCAPED_INDEX)

                if instruction['key'] is None:
                    instruction['key'] = match.group(Grammar.SECTION_KEY_ESCAPED_INDEX)
                    instruction['ranges']['escape_begin_operator'] = match.span(Grammar.SECTION_KEY_ESCAPE_BEGIN_OPERATOR_INDEX)
                    instruction['ranges']['escape_end_operator'] = match.span(Grammar.SECTION_KEY_ESCAPE_END_OPERATOR_INDEX)
                    instruction['ranges']['key'] = match.span(Grammar.SECTION_KEY_ESCAPED_INDEX)
                else:
                    instruction['ranges']['key'] = match.span(Grammar.SECTION_KEY_UNESCAPED_INDEX)

                template = match.group(Grammar.SECTION_TEMPLATE_INDEX)

                if template is not None:
                    instruction['ranges']['template'] = match.span(Grammar.SECTION_TEMPLATE_INDEX)
                    instruction['template'] = template

                    parent = instruction['parent']
                    while parent['type'] != DOCUMENT:
                        parent['deep_resolve'] = True
                        parent = parent['parent']

                    copy_operator_range = match.span(Grammar.SECTION_COPY_OPERATOR_INDEX)

                    if copy_operator_range[1] - copy_operator_range[0] == 2:
                        instruction['deep_copy'] = True
                        instruction['ranges']['deep_copy_operator'] = copy_operator_range
                    else:
                        instruction['ranges']['copy_operator'] = copy_operator_range

                    if not hasattr(self._context, '_copy_sections'):
                        self._context._copy_sections = {}

                    if template in self._context._copy_sections:
                        self._context._copy_sections[template]['targets'].append(instruction)
                    else:
                        self._context._copy_sections[template] = { 'targets': [instruction] }

                    instruction['copy'] = self._context._copy_sections[template]

                last_continuable_element = None
                last_non_section_element = None

            elif match.group(Grammar.MULTILINE_FIELD_OPERATOR_INDEX) is not None:

                if comments is not None:
                    instruction['comments'] = comments
                    comments = None

                operator = match.group(Grammar.MULTILINE_FIELD_OPERATOR_INDEX)
                key = match.group(Grammar.MULTILINE_FIELD_KEY_INDEX)

                instruction['key'] = key
                instruction['parent'] = last_section
                instruction['ranges']['multiline_field_operator'] = match.span(Grammar.MULTILINE_FIELD_OPERATOR_INDEX)
                instruction['ranges']['key'] = match.span(Grammar.MULTILINE_FIELD_KEY_INDEX)
                instruction['type'] = MULTILINE_FIELD_BEGIN

                self._index = match.end()

                terminator_re = re.compile(rf"\n[^\S\n]*({operator})(?!-)[^\S\n]*({re.escape(key)})[^\S\n]*(?=\n|$)")
                terminator_match = terminator_re.search(self._context.input, self._index)

                self._index += 1  # move past current char (\n) into next line
                self._line += 1

                last_section['elements'].append(instruction)
                last_continuable_element = None
                last_non_section_element = instruction

                if not terminator_match:
                    self._tokenize_error_context()
                    raise Parsing.unterminated_multiline_field(self._context, instruction)

                end_of_multiline_field_index = terminator_match.start()

                if end_of_multiline_field_index != self._index - 1:
                    instruction['lines'] = []

                    while True:
                        end_of_line_index = self._context.input.find('\n', self._index, end_of_multiline_field_index)

                        if end_of_line_index == -1:
                            last_non_section_element['lines'].append({
                                'line': self._line,
                                'ranges': {
                                    'line': (self._index, end_of_multiline_field_index),
                                    'value': (self._index, end_of_multiline_field_index)
                                },
                                'type': MULTILINE_FIELD_VALUE
                            })

                            self._index = end_of_multiline_field_index + 1
                            self._line += 1

                            break
                        else:
                            last_non_section_element['lines'].append({
                                'line': self._line,
                                'ranges': {
                                    'line': (self._index, end_of_line_index),
                                    'value': (self._index, end_of_line_index)
                                },
                                'type': MULTILINE_FIELD_VALUE
                            })

                            self._index = end_of_line_index + 1
                            self._line += 1

                instruction = {
                    'line': self._line,
                    'ranges': {
                        'key': terminator_match.span(2),
                        'line': (self._index, terminator_match.end()),
                        'multiline_field_operator': terminator_match.span(1)
                    },
                    'type': MULTILINE_FIELD_END
                }

                last_non_section_element['end'] = instruction
                last_non_section_element = None

                self._index = terminator_match.end() + 1
                self._line += 1

                multiline_field = True

            elif match.group(Grammar.COMMENT_OPERATOR_INDEX) is not None:

                if comments is None:
                    comments = [instruction]
                else:
                    comments.append(instruction)

                instruction['ranges']['comment_operator'] = match.span(Grammar.COMMENT_OPERATOR_INDEX)
                instruction['type'] = COMMENT

                comment = match.group(Grammar.COMMENT_VALUE_INDEX)

                if comment is not None:
                    instruction['comment'] = comment
                    instruction['ranges']['comment'] = match.span(Grammar.COMMENT_VALUE_INDEX)

            elif match.group(Grammar.COPY_OPERATOR_INDEX) is not None:

                if comments is not None:
                    instruction['comments'] = comments
                    comments = None

                template = match.group(Grammar.TEMPLATE_INDEX)

                instruction['ranges']['copy_operator'] = match.span(Grammar.COPY_OPERATOR_INDEX)
                instruction['ranges']['template'] = match.span(Grammar.TEMPLATE_INDEX)
                instruction['template'] = template
                instruction['type'] = EMPTY_ELEMENT

                instruction['key'] = match.group(Grammar.KEY_UNESCAPED_INDEX)

                if instruction['key'] is None:
                    instruction['key'] = match.group(Grammar.KEY_ESCAPED_INDEX)
                    instruction['ranges']['escape_begin_operator'] = match.span(Grammar.KEY_ESCAPE_BEGIN_OPERATOR_INDEX)
                    instruction['ranges']['escape_end_operator'] = match.span(Grammar.KEY_ESCAPE_END_OPERATOR_INDEX)
                    instruction['ranges']['key'] = match.span(Grammar.KEY_ESCAPED_INDEX)
                else:
                    instruction['ranges']['key'] = match.span(Grammar.KEY_UNESCAPED_INDEX)

                instruction['parent'] = last_section
                last_section['elements'].append(instruction)
                last_continuable_element = None
                last_non_section_element = instruction

                if not hasattr(self._context, '_copy_non_section_elements'):
                    self._context._copy_non_section_elements = {}

                if template in self._context._copy_non_section_elements:
                    self._context._copy_non_section_elements[template]['targets'].append(instruction)
                else:
                    self._context._copy_non_section_elements[template] = { 'targets': [instruction] }

                instruction['copy'] = self._context._copy_non_section_elements[template]

            if not multiline_field:
                self._index = match.end() + 1
                self._line += 1

        # again outside while self._index < len(self._context.input):

        if self._context.input[-1] == '\n':
            self._context.line_count = self._line + 1
        else:
            self._context.line_count = self._line

        if comments is not None:
            self._context.meta.extend(comments)



from .error_types import Error, ParseError, ValidationError
from .lookup import lookup
from .parse import parse
from .register import register
from .reporters.html_reporter import HtmlReporter
from .reporters.terminal_reporter import TerminalReporter
from .reporters.text_reporter import TextReporter



from ..constants import BEGIN, END, HUMAN_INDEXING
from .reporter import DISPLAY, EMPHASIZE, INDICATE, OMISSION, QUESTION, Reporter

INDICATORS = {
    DISPLAY: ' ',
    EMPHASIZE: '>',
    INDICATE: '*',
    QUESTION: '?'
}

class TextReporter(Reporter):
    def __init__(self, context):
        super().__init__(context)

        gutter_header = context.messages.gutter_header.rjust(5)
        columns_header = f"  {gutter_header} | {context.messages.content_header}"

        source = f"-- {context.source} --\n\n" if context.source is not None else ''

        self._gutter_width = len(gutter_header) + 3
        self._header = f"{source}{columns_header}\n"

    def _line(self, line, tag):
        if tag == OMISSION:
            return f"{' ' * (self._gutter_width - 5)}..."

        number = str(line + HUMAN_INDEXING)
        instruction = self._index[line]

        if instruction is None:
            content = ''
        else:
            content = self._context.input[instruction['ranges']['line'][BEGIN]:instruction['ranges']['line'][END]]

        return f" {INDICATORS[tag]}{number.rjust(self._gutter_width - 3)} | {content}"

    def _print(self):
        snippet = '\n'.join(self._line(line, tag) for line, tag in enumerate(self._snippet) if tag is not None)

        return self._header + snippet



import re
from ..constants import BEGIN, DOCUMENT, END, HUMAN_INDEXING
from ..error_types import ParseError
from .selections import cursor, select_line, select_template

class Parsing:
    UNTERMINATED_ESCAPED_KEY = re.compile(r'^\s*#*\s*(`+)(?!`)((?:(?!\1).)+)$')

    @staticmethod
    def cyclic_dependency(context, instruction, instruction_chain):
        first_occurrence = instruction_chain.index(instruction)
        feedback_chain = instruction_chain[first_occurrence:]

        if 'template' in feedback_chain[-1]:
            copy_instruction = feedback_chain[-1]
        elif 'template' in feedback_chain[0]:
            copy_instruction = feedback_chain[0]

        reporter = context.reporter(context)

        reporter.report_line(copy_instruction)

        for element in feedback_chain:
            if element != copy_instruction:
                reporter.indicate_line(element)

        return ParseError(
            context.messages.cyclic_dependency(copy_instruction['line'] + HUMAN_INDEXING, copy_instruction['template']),
            reporter.snippet(),
            select_template(copy_instruction)
        )


    @staticmethod
    def invalid_line(context, instruction):
        line = context.input[instruction['ranges']['line'][BEGIN]:instruction['ranges']['line'][END]]

        match = Parsing.UNTERMINATED_ESCAPED_KEY.match(line)
        if match is not None:
            return Parsing.unterminated_escaped_key(context, instruction, match.end(1))

        return ParseError(
            context.messages.invalid_line(instruction['line'] + HUMAN_INDEXING),
            context.reporter(context).report_line(instruction).snippet(),
            select_line(instruction)
        )

    @staticmethod
    def missing_element_for_continuation(context, continuation):
        return ParseError(
            context.messages.missing_element_for_continuation(continuation['line'] + HUMAN_INDEXING),
            context.reporter(context).report_line(continuation).snippet(),
            select_line(continuation)
        )

    @staticmethod
    def missing_fieldset_for_fieldset_entry(context, entry):
        return ParseError(
            context.messages.missing_fieldset_for_fieldset_entry(entry['line'] + HUMAN_INDEXING),
            context.reporter(context).report_line(entry).snippet(),
            select_line(entry)
        )

    @staticmethod
    def missing_list_for_list_item(context, item):
        return ParseError(
            context.messages.missing_list_for_list_item(item['line'] + HUMAN_INDEXING),
            context.reporter(context).report_line(item).snippet(),
            select_line(item)
        )

    @staticmethod
    def non_section_element_not_found(context, copy):
        return ParseError(
            context.messages.non_section_element_not_found(copy['line'] + HUMAN_INDEXING, copy['template']),
            context.reporter(context).report_line(copy).snippet(),
            select_line(copy)
        )

    @staticmethod
    def section_hierarchy_layer_skip(context, section, super_section):
        reporter = context.reporter(context).report_line(section)

        if super_section['type'] != DOCUMENT:
            reporter.indicate_line(super_section)

        return ParseError(
            context.messages.section_hierarchy_layer_skip(section['line'] + HUMAN_INDEXING),
            reporter.snippet(),
            select_line(section)
        )

    @staticmethod
    def section_not_found(context, copy):
        return ParseError(
            context.messages.section_not_found(copy['line'] + HUMAN_INDEXING, copy['template']),
            context.reporter(context).report_line(copy).snippet(),
            select_line(copy)
        )

    @staticmethod
    def unterminated_escaped_key(context, instruction, selection_column):
        return ParseError(
            context.messages.unterminated_escaped_key(instruction['line'] + HUMAN_INDEXING),
            context.reporter(context).report_line(instruction).snippet(),
            {
                'from': { 'column': selection_column, 'index': instruction['ranges']['line'][BEGIN] + selection_column, 'line': instruction['line'] },
                'to': cursor(instruction, 'line', END)
            }
        )

    @staticmethod
    def two_or_more_templates_found(context, copy, first_template, second_template):
        return ParseError(
            context.messages.two_or_more_templates_found(copy['template']),
            context.reporter(context).report_line(copy).question_line(first_template).question_line(second_template).snippet(),
            select_line(copy)
        )

    @staticmethod
    def unterminated_multiline_field(context, field):
        reporter = context.reporter(context).report_element(field)

        # TODO: Possibly revisit this - differs from js impl. and possibly improvable here or there
        for instruction in context.meta:
            if instruction['line'] > field['line']:
                reporter.indicate_line(instruction)

        return ParseError(
            context.messages.unterminated_multiline_field(field['key'], field['line'] + HUMAN_INDEXING),
            reporter.snippet(),
            select_line(field)
        )



from . import fieldset
from .value_element_base import ValueElementBase


class FieldsetEntry(ValueElementBase):
    def __repr__(self):
        return f"<class FieldsetEntry key={self._instruction['key']} value={self._print_value()}>"

    def parent(self):
        return self._parent or fieldset.Fieldset(self._context, self._instruction.parent)



from ..constants import (
    BEGIN,
    END,
    FIELD,
    FIELDSET,
    FIELDSET_ENTRY,
    LIST,
    LIST_ITEM,
    MULTILINE_FIELD_BEGIN,
    SECTION
)

DOCUMENT_BEGIN = {
    'from': { 'column': 0, 'index': 0, 'line': 0 },
    'to': { 'column': 0, 'index': 0, 'line': 0 }
}

def last_in(element):
    if ((element['type'] == FIELD or
         element['type'] == LIST_ITEM or
         element['type'] == FIELDSET_ENTRY) and 'continuations' in element):
        return element['continuations'][-1]
    elif element['type'] == LIST and 'items' in element:
        return last_in(element['items'][-1])
    elif element['type'] == FIELDSET and 'entries' in element:
        return last_in(element['entries'][-1])
    elif element['type'] == MULTILINE_FIELD_BEGIN:
        return element['end']
    elif element['type'] == SECTION and len(element['elements']) > 0:
        return last_in(element['elements'][-1])
    else:
        return element

def cursor(instruction, range, position):
    index = instruction['ranges'][range][position]

    return {
        'column': index - instruction['ranges']['line'][BEGIN],
        'index': index,
        'line': instruction['line']
    }

def selection(from_instruction, from_range, from_position, *to):
    to_instruction = next((argument for argument in to if isinstance(argument, dict)), from_instruction)
    to_range = next((argument for argument in to if isinstance(argument, str)), from_range)
    to_position = next((argument for argument in to if isinstance(argument, int)), from_position)

    return {
        'from': cursor(from_instruction, from_range, from_position),
        'to': cursor(to_instruction, to_range, to_position)
    }

def select_comments(element):
    comments = element['comments']

    if len(comments) == 1:
        if 'comment' in comments[0]:
            return selection(comments[0], 'comment', BEGIN, END)
        else:
            return selection(comments[0], 'line', BEGIN, END)
    elif len(comments) > 1:
        return selection(comments[0], 'line', BEGIN, comments[-1], 'line', END)
    else:
        return selection(element, 'line', BEGIN)

def select_element(element):
    return selection(element, 'line', BEGIN, last_in(element), 'line', END)

def select_key(element):
    return selection(element, 'key', BEGIN, END)

def select_line(element):
    return selection(element, 'line', BEGIN, END)

def select_template(element):
    return selection(element, 'template', BEGIN, END)



from .missing_element_base import MissingElementBase

class MissingList(MissingElementBase):
    def __repr__(self):
        if self._key:
            return f"<class MissingList key={self._key}>"

        return '<class MissingList>'

    def items(self):
        return []

    def optional_string_values(self):
        return []

    def optional_values(self, _loader):
        return []

    def required_string_values(self):
        return []

    def required_values(self, _loader):
        return []



from ..errors.validation import Validation
from . import element as element_module, section_element
from .element_base import ElementBase
from .missing import missing_empty
from .missing import missing_field
from .missing import missing_fieldset
from .missing import missing_list
from .missing import missing_section
from .missing import missing_section_element
from ..constants import (
    DOCUMENT,
    EMPTY_ELEMENT,
    FIELD,
    FIELDSET,
    LIST,
    MULTILINE_FIELD_BEGIN,
    SECTION
)

class Section(ElementBase):
    def __init__(self, context, instruction, parent=None):
        super().__init__(context, instruction, parent)

        self._all_elements_required = parent._all_elements_required if parent else False

    def __repr__(self):
        if self._instruction['type'] == DOCUMENT:
            return f"<class Section document elements={len(self._elements())}>"

        return f"<class Section key=\"{self._instruction['key']}\" elements={len(self._elements())}>"

    def _element(self, key, *, required=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        if len(elements) == 0:
            if required or self._all_elements_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_element')
            elif required is None:
                return missing_section_element.MissingSectionElement(key, self)
            else:
                return None

        if len(elements) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [element._instruction for element in elements],
                'expected_single_element'
            )

        return elements[0]

    def _elements(self, as_map=False):
        if not hasattr(self, '_instantiated_elements'):
            self._instantiated_elements = []
            self._instantiated_elements_map = {}
            self._instantiate_elements(self._instruction)

        return self._instantiated_elements_map if as_map else self._instantiated_elements

    def _empty(self, key, *, required=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        if len(elements) == 0:
            if required or self._all_elements_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_empty')
            elif required is None:
                return missing_empty.MissingEmpty(key, self)
            else:
                return None

        if len(elements) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [element._instruction for element in elements],
                'expected_single_empty'
            )

        element = elements[0]

        if element._instruction['type'] != EMPTY_ELEMENT:
            raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_empty')

        return element.to_empty()

    def _field(self, key, *, required=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        if len(elements) == 0:
            if required or self._all_elements_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_field')
            elif required is None:
                return missing_field.MissingField(key, self)
            else:
                return None

        if len(elements) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [element._instruction for element in elements],
                'expected_single_field'
            )

        element = elements[0]

        if (element._instruction['type'] != FIELD and
            element._instruction['type'] != MULTILINE_FIELD_BEGIN and
            element._instruction['type'] != EMPTY_ELEMENT):
            raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_field')

        return element.to_field()

    def _fieldset(self, key, *, required=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        if len(elements) == 0:
            if required or self._all_elements_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_fieldset')
            elif required is None:
                return missing_fieldset.MissingFieldset(key, self)
            else:
                return None

        if len(elements) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [element._instruction for element in elements],
                'expected_single_fieldset'
            )

        element = elements[0]

        if element._instruction['type'] != FIELDSET and element._instruction['type'] != EMPTY_ELEMENT:
            raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_fieldset')

        return element.to_fieldset()

    def _instantiate_elements(self, section):
        if 'mirror' in section:
            self._instantiate_elements(section['mirror'])
        else:
            def instantiate_and_index(element):
                instance = section_element.SectionElement(self._context, element, self)

                if element['key'] in self._instantiated_elements_map:
                    self._instantiated_elements_map[element['key']].append(instance)
                else:
                    self._instantiated_elements_map[element['key']] = [instance]

                return instance

            filtered = [element for element in section['elements'] if element['key'] not in self._instantiated_elements_map]
            self._instantiated_elements.extend(instantiate_and_index(element) for element in filtered)  # TODO: Probably needs to come AFTER 'if 'extend' in section:' below because otherwise the order is incorrect?

            if 'extend' in section:
                self._instantiate_elements(section['extend'])

    def _list(self, key, *, required=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        if len(elements) == 0:
            if required or self._all_elements_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_list')
            elif required is None:
                return missing_list.MissingList(key, self)
            else:
                return None

        if len(elements) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [element._instruction for element in elements],
                'expected_single_list'
            )

        element = elements[0]

        if element._instruction['type'] != LIST and element._instruction['type'] != EMPTY_ELEMENT:
            raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_list')

        return element.to_list()

    def _missing_error(self, element):
        if isinstance(element, missing_field.MissingField):
            raise Validation.missing_element(self._context, element._key, self._instruction, 'missing_field')
        elif isinstance(element, missing_fieldset.MissingFieldset):
            raise Validation.missing_element(self._context, element._key, self._instruction, 'missing_fieldset')
        elif isinstance(element, missing_list.MissingList):
            raise Validation.missing_element(self._context, element._key, self._instruction, 'missing_list')
        elif isinstance(element, missing_section.MissingSection):
            raise Validation.missing_element(self._context, element._key, self._instruction, 'missing_section')
        else:
            raise Validation.missing_element(self._context, element._key, self._instruction, 'missing_element')

    def _section(self, key, *, required=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        if len(elements) == 0:
            if required or self._all_elements_required:
                raise Validation.missing_element(self._context, key, self._instruction, 'missing_section')
            elif required is None:
                return missing_section.MissingSection(key, self)
            else:
                return None

        if len(elements) > 1:
            raise Validation.unexpected_multiple_elements(
                self._context,
                key,
                [element._instruction for element in elements],
                'expected_single_section'
            )

        element = elements[0]

        if element._instruction['type'] != SECTION:
            raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_section')

        return element.to_section()

    def _untouched(self):
        if not hasattr(self, '_touched'):
            return self._instruction

        for element in self._elements():
            untouched_element = element._untouched()
            if untouched_element:
                return untouched_element

        return False

    def all_elements_required(self, required=True):
        self._all_elements_required = required

        for element in self._elements():
            if element._instruction['type'] == SECTION and element._yielded:
                element.to_section().all_elements_required(required)
            elif element._instruction['type'] == FIELDSET and element._yielded:
                element.to_fieldset().all_entries_required(required)

    def assert_all_touched(self, message=None, *, only=None, skip=None):
        elements_map = self._elements(True)

        for key, elements in elements_map.items():
            if (skip and key in skip) or (only and key not in only):
                continue

            for element in elements:
                untouched = element._untouched()

                if untouched:
                    if callable(message):
                        message = message(element_module.Element(self._context, untouched, self))

                    raise Validation.unexpected_element(self._context, message, untouched)

    def element(self, key=None):
        return self._element(key)

    def elements(self, key=None):
        self._touched = True

        if key:
            elements_map = self._elements(True)
            return elements_map[key] if key in elements_map else []

        return self._elements()

    def empty(self, key=None):
        return self._empty(key)

    def field(self, key=None):
        return self._field(key)

    def fields(self, key=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        def cast(element):
            if (element._instruction['type'] != FIELD and
                element._instruction['type'] != MULTILINE_FIELD_BEGIN and
                element._instruction['type'] != EMPTY_ELEMENT):
                raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_fields')

            return element.to_field()

        return [cast(element) for element in elements]

    def fieldset(self, key=None):
        return self._fieldset(key)

    def fieldsets(self, key=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        def cast(element):
            if element._instruction['type'] != FIELDSET and element._instruction['type'] != EMPTY_ELEMENT:
                raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_fieldsets')

            return element.to_fieldset()

        return [cast(element) for element in elements]

    def list(self, key=None):
        return self._list(key)

    def lists(self, key=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        def cast(element):
            if element._instruction['type'] != LIST and element._instruction['type'] != EMPTY_ELEMENT:
                raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_lists')

            return element.to_list()

        return [cast(element) for element in elements]

    def optional_element(self, key):
        return self._element(key, required=False)

    def optional_empty(self, key):
        return self._empty(key, required=False)

    def optional_field(self, key):
        return self._field(key, required=False)

    def optional_fieldset(self, key):
        return self._fieldset(key, required=False)

    def optional_list(self, key):
        return self._list(key, required=False)

    def optional_section(self, key):
        return self._section(key, required=False)

    def parent(self):
        if self._instruction['type'] == DOCUMENT:
            return None

        return self._parent or Section(self._context, self._instruction['parent'])

    def required_element(self, key=None):
        return self._element(key, required=True)

    def required_empty(self, key=None):
        return self._empty(key, required=True)

    def required_field(self, key=None):
        return self._field(key, required=True)

    def required_fieldset(self, key=None):
        return self._fieldset(key, required=True)

    def required_list(self, key=None):
        return self._list(key, required=True)

    def required_section(self, key=None):
        return self._section(key, required=True)

    def section(self, key=None):
        return self._section(key)

    def sections(self, key=None):
        self._touched = True

        if not key:
            elements = self._elements()
        else:
            elements_map = self._elements(True)
            elements = elements_map[key] if key in elements_map else []

        def cast(element):
            if element._instruction['type'] != SECTION:
                raise Validation.unexpected_element_type(self._context, key, element._instruction, 'expected_sections')

            return element.to_section()

        return [cast(element) for element in elements]

    def touch(self):
        self._touched = True

        for element in self._elements():
            element.touch()



# Note: Study this file from the bottom up

import re


class Grammar:
    OPTIONAL = r'([^\n]+?)?'
    REQUIRED = r'(\S[^\n]*?)'

    #
    EMPTY = '()'
    EMPTY_LINE_INDEX = 1

    # TODO: Here and in other implementations possibly unify CONTINUATION REGEX into one, evaluate type of continuation operator in analyzer then
    # | value
    DIRECT_LINE_CONTINUATION = rf"(\|)[^\S\n]*{OPTIONAL}"
    DIRECT_LINE_CONTINUATION_OPERATOR_INDEX = 2
    DIRECT_LINE_CONTINUATION_VALUE_INDEX = 3

    # \ value
    SPACED_LINE_CONTINUATION = rf"(\\)[^\S\n]*{OPTIONAL}"
    SPACED_LINE_CONTINUATION_OPERATOR_INDEX = 4
    SPACED_LINE_CONTINUATION_VALUE_INDEX = 5

    CONTINUATION = f"{DIRECT_LINE_CONTINUATION}|{SPACED_LINE_CONTINUATION}"

    # > Comment
    COMMENT = rf"(>)[^\S\n]*{OPTIONAL}"
    COMMENT_OPERATOR_INDEX = 6
    COMMENT_VALUE_INDEX = 7

    # - value
    LIST_ITEM = rf"(-)(?!-)[^\S\n]*{OPTIONAL}"
    LIST_ITEM_OPERATOR_INDEX = 8
    LIST_ITEM_VALUE_INDEX = 9

    # -- key
    MULTILINE_FIELD = rf"(-{{2,}})(?!-)[^\S\n]*{REQUIRED}"
    MULTILINE_FIELD_OPERATOR_INDEX = 10
    MULTILINE_FIELD_KEY_INDEX = 11

    # #
    SECTION_OPERATOR = '(#+)(?!#)'
    SECTION_OPERATOR_INDEX = 12

    # # key
    SECTION_KEY_UNESCAPED = r'([^\s`<][^<\n]*?)'
    SECTION_KEY_UNESCAPED_INDEX = 13

    # # `key`
    SECTION_KEY_ESCAPE_BEGIN_OPERATOR_INDEX = 14
    SECTION_KEY_ESCAPED = rf"(`+)(?!`)[^\S\n]*(\S[^\n]*?)[^\S\n]*(\{SECTION_KEY_ESCAPE_BEGIN_OPERATOR_INDEX})" # TODO: Should this exclude the backreference inside the quotes? (as in ((?:(?!\\1).)+) ) here and elsewhere (probably not because it's not greedy.?)
    SECTION_KEY_ESCAPED_INDEX = 15
    SECTION_KEY_ESCAPE_END_OPERATOR_INDEX = 16

    # # key < template
    # # `key` < template
    SECTION_KEY = f"(?:{SECTION_KEY_UNESCAPED}|{SECTION_KEY_ESCAPED})"
    SECTION_TEMPLATE = rf"(?:(<(?!<)|<<)[^\S\n]*{REQUIRED})?"
    SECTION = rf"{SECTION_OPERATOR}\s*{SECTION_KEY}[^\S\n]*{SECTION_TEMPLATE}"
    SECTION_COPY_OPERATOR_INDEX = 17
    SECTION_TEMPLATE_INDEX = 18

    EARLY_DETERMINED = f"{CONTINUATION}|{COMMENT}|{LIST_ITEM}|{MULTILINE_FIELD}|{SECTION}"

    # key
    KEY_UNESCAPED = r'([^\s>#\-`\\|:=<][^\n:=<]*?)'
    KEY_UNESCAPED_INDEX = 19

    # `key`
    KEY_ESCAPE_BEGIN_OPERATOR_INDEX = 20
    KEY_ESCAPED = rf"(`+)(?!`)[^\S\n]*(\S[^\n]*?)[^\S\n]*(\{KEY_ESCAPE_BEGIN_OPERATOR_INDEX})"
    KEY_ESCAPED_INDEX = 21
    KEY_ESCAPE_END_OPERATOR_INDEX = 22

    KEY = f"(?:{KEY_UNESCAPED}|{KEY_ESCAPED})"

    # :
    # : value
    ELEMENT_OR_FIELD = rf"(:)[^\S\n]*{OPTIONAL}"
    ELEMENT_OPERATOR_INDEX = 23
    FIELD_VALUE_INDEX = 24

    # =
    # = value
    FIELDSET_ENTRY = rf"(=)[^\S\n]*{OPTIONAL}"
    FIELDSET_ENTRY_OPERATOR_INDEX = 25
    FIELDSET_ENTRY_VALUE_INDEX = 26

    # < template
    COPY = rf"(<)\s*{REQUIRED}"
    COPY_OPERATOR_INDEX = 27
    TEMPLATE_INDEX = 28

    LATE_DETERMINED = rf"{KEY}\s*(?:{ELEMENT_OR_FIELD}|{FIELDSET_ENTRY}|{COPY})"

    NOT_EMPTY = f"(?:{EARLY_DETERMINED}|{LATE_DETERMINED})"

    REGEX = re.compile(rf"[^\S\n]*(?:{EMPTY}|{NOT_EMPTY})[^\S\n]*(?=\n|$)")



from .missing_element_base import MissingElementBase

class MissingValueElementBase(MissingElementBase):
    def optional_string_value(self):
        return None

    def optional_value(self, _loader):
        return None

    def required_string_value(self):
        self._parent._missing_error(self)

    def required_value(self, _loader):
        self._parent._missing_error(self)



from ..constants import DOCUMENT, FIELDSET_ENTRY, LIST_ITEM
from ..errors.validation import Validation
from . import fieldset_entry
from . import list_item
from .section_element import SectionElement

class Element(SectionElement):
    def __repr__(self):
        return f"<class Element key={self._key()} yields={self._yields()}>"

    def to_document(self):
        if self._instruction['type'] != DOCUMENT:
            raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_document')

        if not hasattr(self, '_section'):
            self._section = section.Section(self._context, self._instruction)
            self._yielded = SECTION

        return self._section

    def to_fieldset_entry(self):
        if not hasattr(self, '_fieldset_entry'):
            if self._instruction['type'] != FIELDSET_ENTRY:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_fieldset_entry')

            self._fieldset_entry = fieldset_entry.FieldsetEntry(self._context, self._instruction)

        return self._fieldset_entry

    def to_list_item(self):
        if not hasattr(self, '_list_item'):
            if self._instruction['type'] != LIST_ITEM:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_list_item')

            self._list_item = list_item.ListItem(self._context, self._instruction)

        return self._list_item

    def to_section(self):
        if not hasattr(self, '_section'):
            if self._instruction['type'] != SECTION and self._instruction['type'] != DOCUMENT:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_section')

            self._section = section.Section(self._context, self._instruction)
            self._yielded = SECTION

        return self._section

    def yields_document(self):
        return self._instruction['type'] == DOCUMENT

    def yields_fieldset_entry(self):
        return self._instruction['type'] == FIELDSET_ENTRY

    def yields_list_item(self):
        return self._instruction['type'] == LIST_ITEM

    def yields_section(self):
        return (self._instruction['type'] == SECTION or
                self._instruction['type'] == DOCUMENT)



from ..error_types import ValidationError
from .selections import cursor, DOCUMENT_BEGIN, select_comments, select_element, select_key, selection
from ..constants import (
    BEGIN,
    END,
    DOCUMENT,
    EMPTY_ELEMENT,
    FIELD,
    FIELDSET_ENTRY,
    LIST_ITEM,
    MULTILINE_FIELD_BEGIN
)


class Validation:
    @staticmethod
    def comment_error(context, message, element):
        return ValidationError(
            context.messages.comment_error(message),
            context.reporter(context).report_comments(element).snippet(),
            select_comments(element)
        )

    @staticmethod
    def element_error(context, message, element):
        return ValidationError(
            message,
            context.reporter(context).report_element(element).snippet(),
            select_element(element)
        )

    @staticmethod
    def key_error(context, message, element):
        return ValidationError(
            context.messages.key_error(message),
            context.reporter(context).report_line(element).snippet(),
            select_key(element)
        )

    @staticmethod
    def missing_comment(context, element):
        return ValidationError(
            context.messages.missing_comment,
            context.reporter(context).report_line(element).snippet(), # TODO: Question-tag an empty line before an element with missing comment
            selection(element, 'line', BEGIN)
        )

    @staticmethod
    def missing_element(context, key, parent, message):
        message = getattr(context.messages, message if key is None else message + '_with_key')

        return ValidationError(
            message if key is None else message(key),
            context.reporter(context).report_missing_element(parent).snippet(),
            DOCUMENT_BEGIN if parent['type'] == DOCUMENT else selection(parent, 'line', END)
        )

    @staticmethod
    def missing_value(context, element):
        selection_data = {}

        if (element['type'] == FIELD or
            element['type'] == EMPTY_ELEMENT or
            element['type'] == MULTILINE_FIELD_BEGIN):
            message = context.messages.missing_field_value(element['key'])

            if 'template' in element['ranges']:
                selection_data['from'] = cursor(element, 'template', END)
            elif 'element_operator' in element['ranges']:
                selection_data['from'] = cursor(element, 'element_operator', END)
            else:
                selection_data['from'] = cursor(element, 'line', END)
        elif element['type'] == FIELDSET_ENTRY:
            message = context.messages.missing_fieldset_entry_value(element['key'])
            selection_data['from'] = cursor(element, 'entry_operator', END)
        elif element['type'] == LIST_ITEM:
            message = context.messages.missing_list_item_value(element['parent']['key'])
            selection_data['from'] = cursor(element, 'item_operator', END)

        snippet = context.reporter(context).report_element(element).snippet()

        if element['type'] == FIELD and 'continuations' in element:
            selection_data['to'] = cursor(element['continuations'][-1], 'line', END)
        else:
            selection_data['to'] = cursor(element, 'line', END)

        return ValidationError(message, snippet, selection_data)

    @staticmethod
    def unexpected_element(context, message, element):
        return ValidationError(
            message or context.messages.unexpected_element,
            context.reporter(context).report_element(element).snippet(),
            select_element(element)
        )

    @staticmethod
    def unexpected_multiple_elements(context, key, elements, message):
        message = getattr(context.messages, message if key is None else message + '_with_key')

        return ValidationError(
            message if key is None else message(key),
            context.reporter(context).report_elements(elements).snippet(),
            select_element(elements[0])
        )

    @staticmethod
    def unexpected_element_type(context, key, section, message):
        message = getattr(context.messages, message if key is None else message + '_with_key')

        return ValidationError(
            message if key is None else message(key),
            context.reporter(context).report_element(section).snippet(),
            select_element(section)
        )

    @staticmethod
    def value_error(context, message, element):
        if 'mirror' in element:
            snippet = context.reporter(context).report_line(element).snippet()
            select = select_key(element)
        elif element['type'] == MULTILINE_FIELD_BEGIN:
            if 'lines' in element:
                snippet = context.reporter(context).report_multiline_value(element).snippet()
                select = selection(element['lines'][0], 'line', BEGIN, element['lines'][-1], 'line', END)
            else:
                snippet = context.reporter(context).report_element(element).snippet()
                select = selection(element, 'line', END)
        else:
            snippet = context.reporter(context).report_element(element).snippet()
            select = {}

            if 'value' in element['ranges']:
                select['from'] = cursor(element, 'value', BEGIN)
            elif 'element_operator' in element['ranges']:
                select['from'] = cursor(element, 'element_operator', END)
            elif 'entry_operator' in element['ranges']:
                select['from'] = cursor(element, 'entry_operator', END)
            elif element['type'] == LIST_ITEM:
                select['from'] = cursor(element, 'item_operator', END)
            else:
                # TODO: Possibly never reached - think through state permutations
                select['from'] = cursor(element, 'line', END)

            if 'continuations' in element:
                select['to'] = cursor(element['continuations'][-1], 'line', END)
            elif 'value' in element['ranges']:
                select['to'] = cursor(element, 'value', END)
            else:
                select['to'] = cursor(element, 'line', END)

        return ValidationError(context.messages.value_error(message), snippet, select)



from setuptools import find_packages, setup

with open('README.md', 'r') as file:
    long_description = file.read()

setup(author='Simon Repp',
      author_email='simon@fdpl.io',
      classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries"
      ],
      description='The eno standard library',
      long_description=long_description,
      long_description_content_type="text/markdown",
      license='MIT',
      name='enolib',
      packages=find_packages(exclude=['performance', 'performance.*', 'tests', 'tests.*']),
      url='https://eno-lang.org/enolib/',
      version='0.7.0',
      zip_safe=False)



from .element_base import ElementBase
from . import section


class Empty(ElementBase):
    def __repr__(self):
        return f"<class Empty key={self._instruction['key']}>"

    def parent(self):
        return self._parent or section.Section(self._context, self._instruction.parent)



from ..constants import BEGIN, COMMENT, END, HUMAN_INDEXING, UNPARSED
from .reporter import DISPLAY, EMPHASIZE, INDICATE, OMISSION, QUESTION, Reporter

RESET = '\x1b[0m'
BOLD = '\x1b[1m'
DIM = '\x1b[2m'

BLACK = '\x1b[30m'
BRIGHT_BLACK = '\x1b[90m'
WHITE = '\x1b[37m'
BRIGHT_WHITE = '\x1b[97m'

BRIGHT_BLACK_BACKGROUND = '\x1b[40m'
BRIGHT_RED_BACKGROUND = '\x1b[101m'
WHITE_BACKGROUND = '\x1b[47m'

INDICATORS = {
    DISPLAY: ' ',
    EMPHASIZE: '>',
    INDICATE: '*',
    QUESTION: '?'
}

GUTTER_STYLE = {
    DISPLAY: BRIGHT_BLACK_BACKGROUND,
    EMPHASIZE: BLACK + BRIGHT_RED_BACKGROUND,
    INDICATE: BLACK + WHITE_BACKGROUND,
    QUESTION: BLACK + WHITE_BACKGROUND
}

RANGE_STYLE = {
    'element_operator': WHITE,
    'escape_begin_operator': WHITE,
    'escape_end_operator': WHITE,
    'item_operator': WHITE,
    'entry_operator': WHITE,
    'section_operator': WHITE,
    'copy_operator': WHITE,
    'deepCopy_operator': WHITE,
    'multiline_field_operator': WHITE,
    'direct_line_continuation_operator': WHITE,
    'spaced_line_continuation_operator': WHITE,
    'key': BOLD + BRIGHT_WHITE,
    'template': BOLD + BRIGHT_WHITE,
    'value': DIM + WHITE
}

class TerminalReporter(Reporter):
    def __init__(self, context):
        super().__init__(context)

        highest_shown_line_number = len(self._snippet)

        for index, tag in reversed(list(enumerate(self._snippet))):
            if tag is not None and tag != OMISSION:
                highest_shown_line_number = index + 1
                break

        self._line_number_padding = max(4, len(str(highest_shown_line_number)))  # TODO: Pick this up in other reporters
        self._header = ''

        if context.source:
            self._header += f"{BLACK + BRIGHT_RED_BACKGROUND} {INDICATORS[EMPHASIZE]} {' '.rjust(self._line_number_padding)} {RESET} {BOLD}{context.source}{RESET}\n"

    def _line(self, line, tag):
        if tag == OMISSION:
            return f"{DIM + BRIGHT_BLACK_BACKGROUND}{'...'.rjust(self._line_number_padding + 2)}  {RESET}"

        number = str(line + HUMAN_INDEXING)
        instruction = self._index[line]

        content = ''
        if instruction:
            if instruction['type'] == COMMENT or instruction['type'] == UNPARSED:
                content = BRIGHT_BLACK + self._context.input[instruction['ranges']['line'][BEGIN]:instruction['ranges']['line'][END]] + RESET
            else:
                content = self._context.input[instruction['ranges']['line'][BEGIN]:instruction['ranges']['line'][END]]

                ranges = [(name, range) for name, range in instruction['ranges'].items() if name != 'line']

                ranges.sort(key=lambda name_range_tuple: name_range_tuple[1][BEGIN], reverse=True)

                for (name, range) in ranges:
                    before = content[:range[BEGIN] - instruction['ranges']['line'][BEGIN]]
                    after = content[range[END] - instruction['ranges']['line'][BEGIN]:]

                    content = before + RANGE_STYLE[name] + self._context.input[range[BEGIN]:range[END]] + RESET + after

        return f"{GUTTER_STYLE[tag]} {INDICATORS[tag]} {number.rjust(self._line_number_padding)} {RESET} {content}"

    def _print(self):
        snippet = '\n'.join(self._line(line, tag) for line, tag in enumerate(self._snippet) if tag is not None)

        return self._header + snippet



from ..errors.validation import Validation
from . import empty
from . import field
from . import fieldset
from . import list as list_module  # don't globally override built-in list function
from . import section
from .element_base import ElementBase
from ..constants import (
    EMPTY_ELEMENT,
    FIELD,
    FIELDSET,
    LIST,
    MULTILINE_FIELD_BEGIN,
    PRETTY_TYPES,
    SECTION
)

class SectionElement(ElementBase):
    def __repr__(self):
        return f"<class SectionElement key={self._key()} yields={self._yields()}>"

    def _yields(self):
        if self._instruction['type'] == EMPTY_ELEMENT:
            return f"{PRETTY_TYPES[EMPTY_ELEMENT]},{PRETTY_TYPES[FIELD]},{PRETTY_TYPES[FIELDSET]},{PRETTY_TYPES[LIST]}"

        return PRETTY_TYPES[self._instruction['type']]

    def _untouched(self):
        if not hasattr(self, '_yielded'):
            return self._instruction

        if hasattr(self, '_empty') and not hasattr(self._empty, '_touched'):
            return self._instruction

        if hasattr(self, '_field') and not hasattr(self._field, '_touched'):
            return self._instruction

        if hasattr(self, '_fieldset'):
            return self._fieldset._untouched()

        if hasattr(self, '_list'):
            return self._list._untouched()

        if hasattr(self, '_section'):
            return self._section._untouched()

    def to_empty(self):
        if not hasattr(self, '_empty'):
            if hasattr(self, '_yielded'):
                raise TypeError(f"This element was already yielded as {PRETTY_TYPES[self._yielded]} and can't be yielded again as an empty.")

            if self._instruction['type'] != EMPTY_ELEMENT:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_empty')

            self._empty = empty.Empty(self._context, self._instruction, self._parent)
            self._yielded = EMPTY_ELEMENT

        return self._empty

    def to_field(self):
        if not hasattr(self, '_field'):
            if hasattr(self, '_yielded'):
                raise TypeError(f"This element was already yielded as {PRETTY_TYPES[self._yielded]} and can't be yielded again as a field.")

            if (self._instruction['type'] != FIELD and
                self._instruction['type'] != MULTILINE_FIELD_BEGIN and
                self._instruction['type'] != EMPTY_ELEMENT):
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_field')

            self._field = field.Field(self._context, self._instruction, self._parent)
            self._yielded = FIELD

        return self._field

    def to_fieldset(self):
        if not hasattr(self, '_fieldset'):
            if hasattr(self, '_yielded'):
                raise TypeError(f"This element was already yielded as {PRETTY_TYPES[self._yielded]} and can't be yielded again as a fieldset.")

            if self._instruction['type'] != FIELDSET and self._instruction['type'] != EMPTY_ELEMENT:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_fieldset')

            self._fieldset = fieldset.Fieldset(self._context, self._instruction, self._parent)
            self._yielded = FIELDSET

        return self._fieldset

    def to_list(self):
        if not hasattr(self, '_list'):
            if hasattr(self, '_yielded'):
                raise TypeError(f"This element was already yielded as {PRETTY_TYPES[self._yielded]} and can't be yielded again as a list.")

            if self._instruction['type'] != LIST and self._instruction['type'] != EMPTY_ELEMENT:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_list')

            self._list = list_module.List(self._context, self._instruction, self._parent)
            self._yielded = LIST

        return self._list

    def to_section(self):
        if not hasattr(self, '_section'):
            if self._instruction['type'] != SECTION:
                raise Validation.unexpected_element_type(self._context, None, self._instruction, 'expected_section')

            self._section = section.Section(self._context, self._instruction, self._parent)
            self._yielded = SECTION

        return self._section

    def touch(self):
        # TODO: Here and elsewhere: This needs to touch anyway; possibly not so small implications
        if not hasattr(self, '_yielded'):
            return

        if hasattr(self, '_empty'):
            self._empty._touched = True

        if hasattr(self, '_field'):
            self._field._touched = True

        if hasattr(self, '_fieldset'):
            self._fieldset.touch()

        if hasattr(self, '_list'):
            self._list.touch()

        if hasattr(self, '_section'):
            self._section.touch()

    def yields_empty(self):
        return self._instruction['type'] == EMPTY_ELEMENT

    def yields_field(self):
        return (self._instruction['type'] == FIELD or
                self._instruction['type'] == MULTILINE_FIELD_BEGIN or
                self._instruction['type'] == EMPTY_ELEMENT)

    def yields_fieldset(self):
        return (self._instruction['type'] == FIELDSET or
                self._instruction['type'] == EMPTY_ELEMENT)

    def yields_list(self):
        return (self._instruction['type'] == LIST or
                self._instruction['type'] == EMPTY_ELEMENT)

    def yields_section(self):
        return self._instruction['type'] == SECTION



from . import list as list_module  # don't globally override built-in list function
from .value_element_base import ValueElementBase


class ListItem(ValueElementBase):
    def __repr__(self):
        return f"<class ListItem value={self._print_value()}>"

    def parent(self):
        return self._parent or list_module.List(self._context, self._instruction.parent)
